#
#  dna : A Library of Procedures for Analysis of DNA Sequences
#  Copyright (C) 2000 J.K. Lindsey
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public Licence as published by
#  the Free Software Foundation; either version 2 of the Licence, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public Licence for more details.
#
#  You should have received a copy of the GNU General Public Licence
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#  SYNOPSIS
#
#	clustalw(dna, names=NULL what="align", gap.open=15, gap.extend=6,
#		quick.pairalign=FALSE, file="phylip.dnd")
#
#  DESCRIPTION
#
#	Function for aligning multiple DNA or protein sequences

clustalw <- function(dna, names=NULL, what="align", gap.open=15, gap.extend=6,
	quick.pairalign=FALSE, file="phylip.dnd"){
tmp <- c("align","align.only","tree.only")
do.what <- match(what <- match.arg(what,tmp),tmp)
if(!is.character(file)||length(strsplit(file,split=NULL)[[1]])<2)
	stop("file must be a file name")
#
# prepare sequences
#
if(is.list(dna)){
	if(is.null(names)){
		if(is.null(names(dna)))stop("names must be supplied")
		else names <- names(dna)}
	dna1 <- ""  # indexing in C code starts at 1, not 0
	for(i in dna)dna1 <- c(dna1,paste(i,collapse=""))}
else if(is.vector(dna)){
	if(is.null(names)){
		if(is.null(names(dna)))stop("names must be supplied")
		else names <- names(dna)}
	dna1 <- c("",dna)}
else if(is.matrix(dna)){
	if(is.null(names)){
		if(is.null(names(dna)))stop("names must be supplied")
		else names <- names(dna)}
	dna1 <- ""
	for(i in 1:dim(dna)[1])
		dna1 <- c(dna1,paste(dna[i,],collapse=""))}
else stop("dna must be a vector, matrix, or list of sequences")
if(!is.character(dna1))stop("dna must contain characters")
seqlen.array <- NULL
for(i in 1:length(dna1))
	seqlen.array <- c(seqlen.array,length(strsplit(dna1[i],split=NULL)[[1]]))
names <- c("",names)
names(dna1) <- names
nseqs <- length(dna1)-1
if(nseqs<2)stop("at least two sequences must be supplied")
#
# call clustalw
#
z <- .C("parse_params",
	do.what=as.integer(do.what),
	nseqs=as.integer(nseqs),
	names=as.character(names),
	seqlen.array=as.integer(seqlen.array),
	seq=as.character(dna1),
	gap.open=as.integer(gap.open),
	gap.extend=as.integer(gap.extend),
	quick.pairalign=as.integer(quick.pairalign),
	file=as.character(file),
	seqout=as.character(rep(paste(rep(" ",2*max(seqlen.array)),collapse=""),nseqs+2)),
	output.index=integer(nseqs+1),
	error=integer(1))
if(z$error<0)switch(-z$error,
	stop("all sequences must have different names"),
	stop("more than one sequence must be supplied"),
	stop("unable to open file"),
	stop("memory allocation problem"),
	stop("matrix maximum resolution is zero"),
	stop("sequences are too divergent to be aligned"),
	stop("unknown residue in matrix"))
#
# return results
#
z$names <- z$names[-1]
z$seqlen.array <- z$seqlen.array[-1]
z$seq <- z$seq[-1]
if(what=="align"||what=="align.only"){
	z$output.index <- z$output.index[-1]
	z$seqout <- matrix(z$seqout[-1],ncol=1)
	dimnames(z$seqout) <- list(c(z$names[z$output.index],"Matches"),"")}
class(z) <- "clustalw"
z
}

print.clustalw <- function(z){
if(z$do.what==1||z$do.what==2){
	cat("\nMultiple sequence alignment:\n")
	print(z$seqout,quote=FALSE)
	if(z$do.what==1){
		cat("\nGap opening penalty:  ",z$gap.open,"\n")
		cat("Gap extending penalty:",z$gap.extend,"\n")}}
}
