#
#  dna : A Library of Procedures for Analysis of DNA Sequences
#  Copyright (C) 2000 J.K. Lindsey
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public Licence as published by
#  the Free Software Foundation; either version 2 of the Licence, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public Licence for more details.
#
#  You should have received a copy of the GNU General Public Licence
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#  SYNOPSIS
#
#	read.DNA(file, skip=0, nlines=0, name="DNA", na.strings=NA,
#		response.obj=FALSE, split=TRUE)
#
#  DESCRIPTION
#
#    Read DNA sequences from a file in standard format

require(rmutil)

read.DNA <- function(file, skip=0, nlines=0, name="DNA", na.strings=NA,
	response.obj=FALSE, split=TRUE){
#
# read in data
#
y <- scan(file,what=character(0),skip=skip,nlines=nlines,quiet=T)
if(response.obj)split <- F
#
# convert to a response object
#
yy <- list()
for(i in seq(2,length(y),by=2)){
	if(split){
		tmp <- strsplit(y[[i]],split=NULL)[[1]]
		if(!is.na(na.strings))for(i in 1:length(na.strings))
			tmp[tmp==na.strings[i]] <- NA}
	else tmp <- y[[i]]
	if(response.obj){
		tmp <- as.numeric(as.factor(tmp))-1
		yy <- c(yy,list(cbind(tmp,1:length(tmp))))}
	else yy <- c(yy,list(tmp))}
if(response.obj)restovec(yy,name=name,type="nominal")
else if(split) yy
else unlist(yy)
}
