#
#  dna : A Library of Procedures for Analysis of DNA Sequences
#  Copyright (C) 2000 J.K. Lindsey
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public Licence as published by
#  the Free Software Foundation; either version 2 of the Licence, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public Licence for more details.
#
#  You should have received a copy of the GNU General Public Licence
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#  SYNOPSIS
#
#	complement.DNA(dna, gaps=NULL)
#	DNAtoProt(dna, mitochondria=FALSE)
#
#  DESCRIPTION
#
#	Utility functions for analyzing DNA sequences

.First.lib <- function(lib, pkg)
	library.dynam("dna", pkg, lib)
require(rmutil)

###
### find complementary sequence
###
complement.DNA <- function(dna, gaps=NULL){
#
# prepare sequence
#
if(!is.vector(dna,mode="character"))
	stop("dna must be a character vector")
s1 <- length(dna)==1
if(s1)dna <- strsplit(dna,split=NULL)[[1]]
rna <- length(grep("U",dna))>0
if(rna)dna <- gsub("U","T",dna)
mat <- match(dna,c(gaps,"A","C","G","T"))
if(any(is.na(mat)))
	stop(paste("dna must only contain A, C, G, T, U, or gaps\nFound",dna[is.na(mat)]))
ldna <- length(dna)
#
# switch codes
#
mA <- match(dna,"A")*(1:ldna)
mC <- match(dna,"C")*(1:ldna)
mG <- match(dna,"G")*(1:ldna)
mT <- match(dna,"T")*(1:ldna)
dna[mA] <- "T"
dna[mC] <- "G"
dna[mG] <- "C"
dna[mT] <- "A"
if(rna)dna <- gsub("T","U",dna)
if(s1)dna <- paste(dna,collapse="")
dna
}

###
### Translate DNA triplets to amino acids
###
DNAtoProt <- function(dna, mitochondria=FALSE){
mp <- function(z) paste(z,collapse="")
#
# prepare vector
#
if(!is.vector(dna,mode="character"))
	stop("dna must be a character vector")
if(length(dna)==1)dna <- strsplit(dna,split=NULL)[[1]]
if(trunc(length(dna)/3)!=length(dna)/3)
	stop("dna must contain complete triplets")
dna <- gsub("T","U",dna)
mat <- match(dna,c("A","C","G","U"))
if(any(is.na(mat)))
	stop(paste("dna must only contain A, C, G, T, or U\nFound",dna[is.na(mat)]))
dna <- apply(matrix(dna,nrow=3),2,mp)
#
# match codes
#
dna <- gsub("UUU|UUC","Phe",dna)
dna <- gsub("UUA|UUG|CUU|CUC|CUG,","Leu",dna)
dna <- gsub("AUU|AUC","Ile",dna)
dna <- gsub("AUA",if(mitochondria)"Met" else "Ile",dna)
dna <- gsub("AUG","Met",dna)
dna <- gsub("GUU|GUC|GUA|GUG","Val",dna)
dna <- gsub("UCU|UCC|UCA|UCG|AGU|AGC","Ser",dna)
dna <- gsub("CCU|CCC|CCA|CCG","Pro",dna)
dna <- gsub("ACU|ACC|ACA|ACG","Thr",dna)
dna <- gsub("GCU|GCC|GCA|GCG","Ala",dna)
dna <- gsub("UAU|UAC","Tyr",dna)
dna <- gsub("UAA|UAG","STP",dna)
dna <- gsub("UGA",if(mitochondria)"Trp" else "STP",dna)
dna <- gsub("CAU|CAC","His",dna)
dna <- gsub("CAA|CAG","Gln",dna)
dna <- gsub("AAU|AAC","Asn",dna)
dna <- gsub("AAA|AAG","Lys",dna)
dna <- gsub("GAU|GAC","Asp",dna)
dna <- gsub("GAA|GAG","Glu",dna)
dna <- gsub("UGU|UGC","Cys",dna)
dna <- gsub("UGG","Trp",dna)
dna <- gsub("CGU|CGC|CGA|CGG|AGA|AGG","Arg",dna)
dna <- gsub("GGU|GGC|GGA|GGG","Gly",dna)
dna
}
