\name{mutations}
\title{Calculate Frequencies of Mutation in a Set of Sequences}
\alias{mutations}
\usage{
mutations(dna, gaps=NULL, group=FALSE)
}
\description{
\code{mutations} calculates the frequency of each base and of a gap at
each site in a set of DNA sequences. All sequences must have the same
length. The bases must be coded as \code{A}, \code{C}, \code{G}, and
\code{T} in capital letters if a matrix is supplied.
}
\arguments{
\item{dna}{A matrix containing two or more sequences as rows, a
list of two vectors possibly of unequal length, or a \code{response}
or \code{repeated} data object containing two or more sequences.}
\item{gaps}{The code(s) for gaps. If none is provided, frequencies of
gaps are not included in the contingency table.}
\item{group}{If FALSE, the frequencies for all bases and gaps are
returned; otherwise, the frequency of the most common base and of all
others grouped together at each site is returned.}
}
\value{
Returns a two-way contingency table containing either the frequencies
of the four bases (and gaps) at each site or the frequency of the most
common base and the sum of all others at each site.
}
\seealso{
\code{\link[dna]{dotmatrix}}, \code{\link[dna]{read.DNA}}
}
\author{J.K. Lindsey}
\examples{
sequences <- rbind(
	c("T","T","C","T","A","T","C","A","T","G","G","G","G","A","A",
	"G","C","A","G","A","T","T","T","G","G","G","T","A","C","C",
	"A","C","C","C","A","A","G","T","A","T","T","G","A","C","T"),
	c("T","T","C","T","T","T","C","A","T","G","G","G","G","A","T",
	"G","C","A","G","A","T","T","T","G","G","G","T","A","C","C",
	"A","G","C","C","A","A","G","T","A","T","T","G","A","C","A"),
	c("T","T","C","T","T","T","C","A","T","G","G","G","G","A","T",
	"G","C","A","G","A","T","T","T","G","G","G","T","A","C","C",
	"A","G","C","C","A","A","G","T","A","-","T","G","A","C","A"),
	c("T","T","C","T","T","T","C","A","T","G","G","G","G","A","T",
	"G","C","A","G","A","T","T","T","G","G","G","T","A","C","C",
	"A","G","C","C","A","A","G","T","A","T","T","G","A","C","A"))
mutations(sequences)
mutations(sequences,gap="-")
mutations(sequences,group=TRUE)
mutations(sequences,group=TRUE,gap="-")
}
\keyword{manip}
