/*
  dna : A Library of Procedures for Analysis of DNA Sequences
  Copyright (C) 2000 J.K. Lindsey

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  DESCRIPTION

	Function for aligning multiple DNA or protein sequences
*/
/* command line interface for Clustal W  */
/* DES was here MARCH. 1994 */
/* DES was here SEPT.  1994 */
/* modified to be R interface, March, 2000, J.K. Lindsey */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <signal.h>
#include <setjmp.h>
#include "clustalw.h"
#include "param.h"

/*
*	Prototypes
*/

static void reset_align(void);
void print_sec_struct_mask(int prf_length,char *mask,char *struct_mask);

/*
*	 Global variables
*/

int max_names;
double  **tmat;
float    gap_open,      gap_extend;
float  	dna_gap_open,  dna_gap_extend;
float 	prot_gap_open, prot_gap_extend;
float    pw_go_penalty,      pw_ge_penalty;
float  	dna_pw_go_penalty,  dna_pw_ge_penalty;
float 	prot_pw_go_penalty, prot_pw_ge_penalty;
sint    wind_gap,ktup,window,signif;
sint    dna_wind_gap, dna_ktup, dna_window, dna_signif;
sint    prot_wind_gap,prot_ktup,prot_window,prot_signif;
short usermat[NUMRES][NUMRES], pw_usermat[NUMRES][NUMRES];
short def_aa_xref[NUMRES+1], aa_xref[NUMRES+1], pw_aa_xref[NUMRES+1];
short userdnamat[NUMRES][NUMRES], pw_userdnamat[NUMRES][NUMRES];
short def_dna_xref[NUMRES+1], dna_xref[NUMRES+1], pw_dna_xref[NUMRES+1];
sint    nseqs;
sint nsets;
sint	new_seq;
sint 	*seqlen_array;
sint 	divergence_cutoff;
sint **sets;
Boolean 	neg_matrix;
Boolean  quick_pairalign;
Boolean	reset_alignments_new;		/* DES */
Boolean	reset_alignments_all;		/* DES */
sint 	gap_dist;
Boolean 	no_hyd_penalties, no_pref_penalties;
sint 	max_aa;
sint 	gap_pos1, gap_pos2;
sint  	max_aln_length;
sint 	*output_index, output_order;
sint     profile_no;
sint	*seq_weight;
sint max_aa;
sint gap_pos1;
sint gap_pos2;
sint mat_avscore;
sint profile_no;

Boolean 	lowercase; /* Flag for GDE output - set on comm. line*/
Boolean 	cl_seq_numbers;
Boolean 	output_clustal, output_nbrf, output_phylip, output_gcg, output_gde;
Boolean 	output_tree_clustal, output_tree_phylip, output_tree_distances;
Boolean 	tossgaps, kimura;
Boolean  percent;
Boolean 	explicit_dnaflag;  /* Explicit setting of sequence type on comm.line*/
Boolean 	showaln, save_parameters;
Boolean	dnaflag;
Boolean distance_tree;
float	transition_weight;

char 	hyd_residues[];
char 	*amino_acid_codes;
char 	*paramstr;

char 	**seq_array;
char 	**names;

char *gap_penalty_mask1,*gap_penalty_mask2;
char *sec_struct_mask1,*sec_struct_mask2;
sint struct_penalties,struct_penalties1,struct_penalties2;
sint output_struct_penalties;
Boolean use_ss1, use_ss2;

char *sec_struct_mask = NULL;
char *gap_penalty_mask = NULL;

Boolean empty;

char 	phylip_tree_name[FILENAMELEN+1]="";

static char *params[MAXARGS];
static char *param_arg[MAXARGS];

static sint numparams;
static Boolean check_tree = TRUE;

sint 	profile1_nseqs;	/* have been filled; the no. of seqs in prof 1*/

static char *lin2;

Boolean user_series = FALSE;
UserMatSeries matseries;
short usermatseries[MAXMAT][NUMRES][NUMRES];
short aa_xrefseries[MAXMAT][NUMRES+1];

void parse_params(int *do_what, int *nseqsr, char *namesr[],
		  int seqlen_arrayr[], char **seq, int *gap_openr,
		  int *gap_extendr, int *quick_pair, char **filer,
		  char **seqout, int *output_indexr, int *error)
{
	sint i,j,len,temp;

	nseqs=*nseqsr;
	names=ckalloc((nseqs+1)*sizeof(sint));
	for(i=1;i<=nseqs;i++)names[i]=namesr[i];
	seqlen_array=seqlen_arrayr;
	output_index=output_indexr;
	quick_pairalign=*quick_pair;
	strcpy(phylip_tree_name,filer[0]);

	max_aln_length = 0;
	profile_no = 0;

	init_matrix(error);
	if(*error<0)return;
	fill_chartab();

	if(*error = readseqs(nseqs, names, seqlen_array, seq)<0)
	  return;

/****************************************************************************/
/* Now do whatever has been requested ***************************************/
/****************************************************************************/
	switch(*do_what){
	case 1:	
	  align(phylip_tree_name,seqout,gap_openr,gap_extendr,error);
	  if(*error<0)return;
	  break;
	case 2:
	  get_tree(phylip_tree_name,seqout,gap_openr,gap_extendr,error);
	  break;
	case 3:
	  make_tree(phylip_tree_name,error);
	  if(*error<0)return;
	  break;}

        for(i=0;i<nseqs;i++)free_seq(i); 
        free_aln(nseqs);
	ckfree(names);
/*******whew!***now*go*home****/
}

void align(char *phylip_name, char **seqout, int *gap_openr,
	   int *gap_extendr, int *error)
{
	FILE *tree;
	sint count, length, i;

	struct_penalties1 = struct_penalties2 = NONE;

	if (nseqs >= 2) {
	  if((tree = open_explicit_file(phylip_name))==NULL){
	    *error=-3;
	    return;}
	}

        if(dnaflag) {
                if(*gap_openr>0)gap_open = *gap_openr;
		else gap_open = *gap_openr = dna_gap_open;
                if(*gap_extendr>0)gap_extend = *gap_extendr;
		else gap_extend = *gap_extendr = dna_gap_extend;
                pw_go_penalty  = dna_pw_go_penalty;
                pw_ge_penalty  = dna_pw_ge_penalty;
                ktup       = dna_ktup;
                window     = dna_window;
                signif     = dna_signif;
                wind_gap   = dna_wind_gap;
        }
        else {
                if(*gap_openr>0)gap_open   = prot_gap_open;
		else gap_open = *gap_openr = prot_gap_open;
                if(*gap_extendr>0)gap_extend = prot_gap_extend;
		else gap_extend = *gap_extendr = prot_gap_extend;
                pw_go_penalty  = prot_pw_go_penalty;
                pw_ge_penalty  = prot_pw_ge_penalty;
                ktup       = prot_ktup;
                window     = prot_window;
                signif     = prot_signif;
                wind_gap   = prot_wind_gap;
        }

        if (quick_pairalign){
	  show_pair((sint)0,nseqs,(sint)0,nseqs,error);
	  if(*error<0)return;}
        else
	  if((*error=pairalign((sint)0,nseqs,(sint)0,nseqs))<0)return;

	if (nseqs >= 2) {
	  guide_tree(tree,1,nseqs,error);
	  if(*error<0)return;
	}
	count = malign((sint)0,phylip_name);
	if (count <= 0){
	  *error=count;
	  return;}
	length=0;
        for (i=1;i<=nseqs;i++)
	  if (length < seqlen_array[i])length = seqlen_array[i];
	clustal_out(seqout, 1, length, 1, nseqs);
}

FILE *  open_explicit_file(char *file_name)
{ 
	FILE * file_handle;

	if (*file_name == EOS) {
		return NULL;
	}
	if((file_handle=fopen(file_name,"w"))==NULL) {
		return NULL;
	}
	return file_handle;
}

void print_sec_struct_mask(int prf_length, char *mask, char *struct_mask)
{
	int i,j;

/*
    calculate the gap penalty mask from the secondary structures
*/
	i=0;
	while (i<prf_length) {
		if (tolower(mask[i]) == 'a' || mask[i] == '$') {
			for (j = 0; j<helix_end_minus; j++) {
				if (i+j>=prf_length || (tolower(mask[i+j]) != 'a'
				                    && mask[i+j] != '$')) break;
				struct_mask[i+j] = 'a';
			}
			i += j;
			while (tolower(mask[i]) == 'a'
				                    || mask[i] == '$') {
				if (i>=prf_length) break;
				if (mask[i] == '$') {
					struct_mask[i] = 'A';
					i++;
					break;
				}
				else struct_mask[i] = mask[i];
				i++;
			}
			for (j = 0; j<helix_end_minus; j++) {
				if ((i-j-1>=0) && (tolower(mask[i-j-1]) == 'a'
				                    || mask[i-j-1] == '$'))
					struct_mask[i-j-1] = 'a';
			}
		}
	 	else if (tolower(mask[i]) == 'b' || mask[i] == '%') {
			for (j = 0; j<strand_end_minus; j++) {
				if (i+j>=prf_length || (tolower(mask[i+j]) != 'b'
				                    && mask[i+j] != '%')) break;
				struct_mask[i+j] = 'b';
			}
			i += j;
			while (tolower(mask[i]) == 'b'
				                    || mask[i] == '%') {
				if (i>=prf_length) break;
				if (mask[i] == '%') {
					struct_mask[i] = 'B';
					i++;
					break;
				}
				else struct_mask[i] = mask[i];
				i++;
			}
			for (j = 0; j<strand_end_minus; j++) {
				if ((i-j-1>=0) && (tolower(mask[i-j-1]) == 'b'
				                    || mask[i-j-1] == '%'))
				struct_mask[i-j-1] = 'b';
			}
		}
	else i++;
	}
}

void clustal_out(char **clusout, sint fres, sint len, sint fseq, sint lseq)
{
	static char *seq1;
	static sint *seq_no;
	static sint *print_seq_no;
	char *ss_mask1, *ss_mask2;
	char c;
	sint val;
	sint ii,lv1,catident1[NUMRES],catident2[NUMRES],ident,chunks;
	sint i,j,k,l;
	sint pos,ptr;
	sint line_length;

	seq_no = (sint *)ckalloc((nseqs+1) * sizeof(sint));
	print_seq_no = (sint *)ckalloc((nseqs+1) * sizeof(sint));
	for (i=fseq;i<=lseq;i++)
        {
		print_seq_no[i] = seq_no[i] = 0;
                for(j=1;j<fres;j++) {
			val = seq_array[i][j];
			if((val >=0) || (val <=max_aa)) seq_no[i]++;
		}
        }

	seq1 = (char *)ckalloc((max_aln_length+1) * sizeof(char));

	if (struct_penalties1 == SECST && use_ss1 == TRUE) {
		ss_mask1 = (char *)ckalloc((seqlen_array[1]+10) * sizeof(char));
		for (i=0;i<seqlen_array[1];i++)
			ss_mask1[i] = sec_struct_mask1[i];
		print_sec_struct_mask(seqlen_array[1],sec_struct_mask1,ss_mask1);
	}
	if (struct_penalties2 == SECST && use_ss2 == TRUE) {
		ss_mask2 = (char *)ckalloc((seqlen_array[profile1_nseqs+1]+10) * sizeof(char));
		for (i=0;i<seqlen_array[profile1_nseqs+1];i++)
			ss_mask2[i] = sec_struct_mask2[i];
		print_sec_struct_mask(seqlen_array[profile1_nseqs+1],sec_struct_mask2,ss_mask2);
	}

/* decide the line length for this alignment - maximum is LINELENGTH */

	line_length=len;
	chunks = 1;
		
	pos = 1;
	ptr = len;
	
	for(ii=fseq;ii<=lseq;++ii) {
                i=output_index[ii];
		print_seq_no[i] = 0;
		for(j=pos;j<=ptr;++j) {
			if (j+fres-1<=seqlen_array[i])
				val = seq_array[i][j+fres-1];
			else val = -3;
			if((val == -3) || (val == 253)) break;
			else if((val < 0) || (val > max_aa))
                                seq1[j]='-';
			else {
				seq1[j]=amino_acid_codes[val];
				seq_no[i]++;
				print_seq_no[i]=1;
			}
		}
		for(;j<=ptr;++j) seq1[j]='-';
		strncpy(clusout[ii],&seq1[pos],ptr-pos+1);
		clusout[ii][ptr-pos+1]=EOS;
	}
	
	for(i=pos;i<=ptr;++i) {
		seq1[i]=' ';
		ident=0;
		for(j=1;res_cat1[j-1]!=NULL;j++) catident1[j-1] = 0;
		for(j=1;res_cat2[j-1]!=NULL;j++) catident2[j-1] = 0;
		for(j=fseq;j<=lseq;++j) {
			if((seq_array[fseq][i] >=0) && 
			   (seq_array[fseq][i] <= max_aa)) {
				if(seq_array[fseq][i] == seq_array[j][i])
				++ident;
				for(k=1;res_cat1[k-1]!=NULL;k++) {
				        for(l=0;(c=res_cat1[k-1][l]);l++) {
						if (amino_acid_codes[seq_array[j][i]]==c)
						{
							catident1[k-1]++;
							break;
						}
					}
				}
				for(k=1;res_cat2[k-1]!=NULL;k++) {
				        for(l=0;(c=res_cat2[k-1][l]);l++) {
						if (amino_acid_codes[seq_array[j][i]]==c)
						{
							catident2[k-1]++;
							break;
						}
					}
				}
			}
		}
		if(ident==lseq-fseq+1)
			seq1[i]='*';
		else if (!dnaflag) {
			for(k=1;res_cat1[k-1]!=NULL;k++) {
				if (catident1[k-1]==lseq-fseq+1) {
					seq1[i]=':';
					break;
				}
			}
			if(seq1[i]==' ')
				for(k=1;res_cat2[k-1]!=NULL;k++) {
					if (catident2[k-1]==lseq-fseq+1) {
						seq1[i]='.';
						break;
					}
				}
		}
	}
	strncpy(clusout[lseq+1],&seq1[pos],ptr-pos+1);
	clusout[lseq+1][ptr-pos+1]=EOS;

	seq1=ckfree((void *)seq1);
	if (struct_penalties1 == SECST && use_ss1 == TRUE) ckfree(ss_mask1);
	if (struct_penalties2 == SECST && use_ss2 == TRUE) ckfree(ss_mask2);

}

void make_tree(char *phylip_name, int *error)
{
	char path[FILENAMELEN+1];
	FILE *tree;

	   struct_penalties1 = struct_penalties2 = NONE;
	   if (sec_struct_mask1 != NULL) sec_struct_mask1=ckfree(sec_struct_mask1);
	   if (sec_struct_mask2 != NULL) sec_struct_mask2=ckfree(sec_struct_mask2);
	   if (gap_penalty_mask1 != NULL) gap_penalty_mask1=ckfree(gap_penalty_mask1);
	   if (gap_penalty_mask2 != NULL) gap_penalty_mask2=ckfree(gap_penalty_mask2);

	if(reset_alignments_new || reset_alignments_all) reset_align();

        if(dnaflag) {
                gap_open   = dna_gap_open;
                gap_extend = dna_gap_extend;
                pw_go_penalty  = dna_pw_go_penalty;
                pw_ge_penalty  = dna_pw_ge_penalty;
                ktup       = dna_ktup;
                window     = dna_window;
                signif     = dna_signif;
                wind_gap   = dna_wind_gap;

        }
        else {
                gap_open   = prot_gap_open;
                gap_extend = prot_gap_extend;
                pw_go_penalty  = prot_pw_go_penalty;
                pw_ge_penalty  = prot_pw_ge_penalty;
                ktup       = prot_ktup;
                window     = prot_window;
                signif     = prot_signif;
                wind_gap   = prot_wind_gap;

        }
   
        if (quick_pairalign){
	  show_pair((sint)0,nseqs,(sint)0,nseqs,error);
	  if(*error<0)return;}
        else
	  if((*error=pairalign((sint)0,nseqs,(sint)0,nseqs))<0)return;

	if (nseqs >= 2) {
	  if((tree = open_explicit_file(phylip_name))==NULL) return;

	  guide_tree(tree,1,nseqs,error);
	  if(*error<0)return;
	}
	
	if(reset_alignments_new || reset_alignments_all) reset_align();

}

static void reset_align(void)   /* remove gaps from older alignments (code =
gap_pos1) */
{		      				/* EXCEPT for gaps that were INPUT with the seqs.*/
	register sint sl;   		     /* which have  code = gap_pos2  */
	sint i,j;
	
	for(i=1;i<=nseqs;++i) {
		sl=0;
		for(j=1;j<=seqlen_array[i];++j) {
			if(seq_array[i][j] == gap_pos1 && (reset_alignments_new ||
reset_alignments_all)) continue;
			if(seq_array[i][j] == gap_pos2 && (reset_alignments_all)) continue;
			++sl;
			seq_array[i][sl]=seq_array[i][j];
		}
		seqlen_array[i]=sl;
	}
	
			
}

void get_tree(char *phylip_name, char **seqout, int *gap_openr,
	   int *gap_extendr, int *error)
{
	FILE *tree;
	sint count, i, length;

	   struct_penalties1 = struct_penalties2 = NONE;
	   if (sec_struct_mask1 != NULL) sec_struct_mask1=ckfree(sec_struct_mask1);
	   if (sec_struct_mask2 != NULL) sec_struct_mask2=ckfree(sec_struct_mask2);
	   if (gap_penalty_mask1 != NULL) gap_penalty_mask1=ckfree(gap_penalty_mask1);
	   if (gap_penalty_mask2 != NULL) gap_penalty_mask2=ckfree(gap_penalty_mask2);

	if(reset_alignments_new || reset_alignments_all) reset_align();

        if (nseqs >= 2) {

        	if((tree=fopen(phylip_name,"r"))==NULL) {
		  *error=-3;
                	return;
        	}
	}
	else {
        	if(dnaflag) {
		  if(*gap_openr>0)gap_open = *gap_openr;
		  else gap_open = *gap_openr = dna_gap_open;
		  if(*gap_extendr>0)gap_extend = *gap_extendr;
		  else gap_extend = *gap_extendr = dna_gap_extend;
                	pw_go_penalty  = dna_pw_go_penalty;
                	pw_ge_penalty  = dna_pw_ge_penalty;
                	ktup       = dna_ktup;
                	window     = dna_window;
                	signif     = dna_signif;
                	wind_gap   = dna_wind_gap;

        	}
        	else {
		  if(*gap_openr>0)gap_open   = prot_gap_open;
		  else gap_open = *gap_openr = prot_gap_open;
		  if(*gap_extendr>0)gap_extend = prot_gap_extend;
		  else gap_extend = *gap_extendr = prot_gap_extend;
                	pw_go_penalty  = prot_pw_go_penalty;
                	pw_ge_penalty  = prot_pw_ge_penalty;
                	ktup       = prot_ktup;
                	window     = prot_window;
                	signif     = prot_signif;
                	wind_gap   = prot_wind_gap;

        	}

            if (quick_pairalign){
	      show_pair((sint)0,nseqs,(sint)0,nseqs,error);
	      if(*error<0)return;}
            else
	      if((*error=pairalign((sint)0,nseqs,(sint)0,nseqs))<0)return;
	}

	count = malign(0,phylip_name);
	if (count <= 0){
	  *error=count;
	  return;}
	length=0;
        for (i=1;i<=nseqs;i++)
	  if (length < seqlen_array[i])length = seqlen_array[i];
	clustal_out(seqout, 1, length, 1, nseqs);

}
