#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>
#include "clustalw.h"

extern char **seq_array;
extern sint *seq_weight;
extern double **tmat;


/*
*	ckalloc()
*
*	Tries to allocate "bytes" bytes of memory. Exits program if failed.
*	Return value:
*		Generic pointer to the newly allocated memory.
*/

void *ckalloc(size_t bytes)
{
	register void *ret;
	
	ret = calloc(bytes, sizeof(char)); /*
	if( (ret = calloc(bytes, sizeof(char))) == NULL)
		fatal("Out of memory\n");
	else
		return ret;	*/

	return ret;	
}

/*
*	ckrealloc()
*
*	Tries to reallocate "bytes" bytes of memory. Exits program if failed.
*	Return value:
*		Generic pointer to the re-allocated memory.
*/

void *ckrealloc(void *ptr, size_t bytes)
{
	register void *ret=NULL;

	ret = realloc(ptr, bytes); /*
	if (ptr == NULL)	
		fatal("Bad call to ckrealloc\n");
	else if( (ret = realloc(ptr, bytes)) == NULL)
		fatal("Out of memory\n");
	else
	return ret;	*/

	return ret;	
}

/*
*	ckfree()
*
*	Tries to free memory allocated by ckalloc.
*	Return value:
*		None.
*/

void *ckfree(void *ptr)
{
  if (ptr != NULL){
	 	free(ptr);
		ptr = NULL;
	}
	return ptr;
}

void alloc_aln(sint nseqs)
{
	sint i;
        seq_array = (char **)ckalloc( (nseqs + 1) * sizeof (char *) );
        tmat = (double **) ckalloc( (nseqs+1) * sizeof (double *) );
        for(i=1;i<=nseqs;i++)
                tmat[i] = (double *)ckalloc( (nseqs+1) * sizeof (double) );

        seq_weight = (sint *)ckalloc( (nseqs+1) * sizeof (sint));
        for(i=1;i<=nseqs;i++)
		seq_weight[i]=100;
}

void realloc_aln(sint first_seq,sint nseqs)
{
	sint i;

        seq_array = (char **)ckrealloc(seq_array, (first_seq+nseqs+1) * sizeof (char *) );

        seq_weight = (sint *)ckrealloc(seq_weight, (first_seq+nseqs+1) * sizeof (sint));
        for(i=first_seq;i<first_seq+nseqs;i++)
		seq_weight[i]=100;

        tmat = (double **) ckrealloc(tmat, (first_seq+nseqs+1) * sizeof (double *) );
        for(i=1;i<first_seq;i++)
                tmat[i] = (double *)ckrealloc(tmat[i], (first_seq+nseqs+1) * sizeof (double) );
        for(i=first_seq;i<first_seq+nseqs;i++)
                tmat[i] = (double *)ckalloc( (first_seq+nseqs+1) * sizeof (double) );
}

void free_aln(sint nseqs)
{
	sint i;

	if(nseqs<=0) return;
        for(i=1;i<=nseqs;i++)
		seq_array[i] = ckfree(seq_array[i]);
	seq_array = ckfree(seq_array);

	seq_weight = ckfree(seq_weight);

        for(i=1;i<=nseqs;i++)
		tmat[i] = ckfree(tmat[i]);
	tmat = ckfree(tmat);
}

void alloc_seq(sint seq_no,sint length)
{
	seq_array[seq_no] = (char *)ckalloc((length+2) * sizeof (char));
}

void realloc_seq(sint seq_no,sint length)
{
	seq_array[seq_no] = (char *)realloc(seq_array[seq_no], (length+2) * sizeof (char));

}

void free_seq(sint seq_no)
{
	seq_array[seq_no]=ckfree(seq_array[seq_no]);
}

